# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import cast
from typing import overload
from typing import Any
from typing import AsyncContextManager
from typing import AsyncIterable
from typing import Generic
from typing import Iterable
from typing import Protocol
from typing import TypeVar

import aiofiles
import pydantic
import yaml


from canonical import TypedResourceName
from canonical.protocols import IStorage
from canonical.exceptions import DoesNotExist
from canonical.exceptions import MultipleObjectsReturned
from canonical.exceptions import Stale
from .iresourcequery import IResourceQuery
from .resource import Resource
from .rootresource import ResourceType
from .rootresource import RootResource


R = TypeVar('R', bound=Resource[Any] | RootResource[Any])
K = TypeVar('K')
T = TypeVar('T', bound=pydantic.BaseModel)


class IResourceRepository(IStorage[ResourceType], Protocol):
    DoesNotExist = DoesNotExist
    MultipleObjectsReturned = MultipleObjectsReturned
    Stale = Stale

    def typed(self, model: type[R]) -> 'TypedResourceRepository[R]':
        return TypedResourceRepository(
            repo=self,
            model=model
        )

    def all(self, model: type[R], namespace: str | None = None) -> AsyncIterable[R]:
        ...

    async def allocate(self, obj: type[ResourceType]) -> int:
        ...

    async def get_by_name(self, model: type[R], name: str | int, namespace: str | None = None) -> R:
        ...

    async def import_documents(
        self,
        kind: Any,
        path: str,
        iterator: bool = True
    ) -> AsyncIterable[Resource[Any] | RootResource[Any]]:
        adapter = pydantic.TypeAdapter(kind)
        async with aiofiles.open(path) as f: # type: ignore
            documents: list[dict[str, Any]] = yaml.safe_load_all(await f.read()) # type: ignore
        async with self.transaction() as tx:
            for doc in documents:
                resource = cast(Resource[Any], adapter.validate_python(doc))
                await resource.persist(self, transaction=tx) # type: ignore
                yield resource

    def query(
        self,
        model: type[T],
        filters: Iterable[tuple[str, str, Any]] | None = None,
        sort: Iterable[str] | None = None,
        namespace: str | None = None,
        limit: int | None = None,
        kind: str | None = None,
        page_size: int = 10,
        keys: list[Any] | None = None,
        **kwargs: Any
    ) -> IResourceQuery[T]:
        ...


    def transaction(self, transaction: Any = None) -> AsyncContextManager[Any]:
        ...


class TypedResourceRepository(Generic[R]):
    model: type[R]

    def __init__(self, repo: IResourceRepository, model: type[R]):
        self.model = model
        self.repo = repo

    @overload
    async def get(self, d: TypedResourceName[R], namespace: str | None) -> R:
        ...

    @overload
    async def get(self, d: int | str, namespace: str | None) -> R:
        ...

    async def get(self, d: Any, *args: Any, **kwargs: Any) -> Any:
        if isinstance(d, TypedResourceName):
            return await self.repo.get_by_name(cast(TypedResourceName[R], d).model, d.id)
        elif inspect.isclass(d) and issubclass(d, (Resource, RootResource)):
            raise NotImplementedError
        elif isinstance(d, (int, str)):
            return await self.repo.get_by_name(self.model, d)
        else:
            raise NotImplementedError(f"Unsupported object type: {type(d).__name__}")


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
import logging

from canonical.ext.iam import BaseAuthorizationContext
from canonical.ext.iam.types import PermissionSet
from canonical.ext.resource import ResourceType
from ..resourceclient import ResourceClient
from .impersonationauth import ImpersonationAuth
from .requestemail import RequestEmail
from .resourcerepository import ResourceRepository


__all__: list[str] = [
    'RemoteAuthorizationContext'
]


class RemoteAuthorizationContext(BaseAuthorizationContext):
    model: type[ResourceType]
    logger: logging.Logger = logging.getLogger('uvicorn')
    verb: str

    @property
    def client(self) -> ResourceClient:
        return ResourceClient(
            auth=self.auth,
            base_url='https://core.molanoapis.com',
            #base_url='https://127.0.0.1:8001'
        )

    def __init__(
        self,
        request: fastapi.Request,
        auth: ImpersonationAuth,
        resources: ResourceRepository,
        email: RequestEmail,
    ):
        self.auth = auth
        self.email = email
        self.granted = PermissionSet()
        self.repo = resources
        self.request = request
        self.subject_type = 'User'

    async def setup(self):
        self.logger.info("Initializing authorization context (subject: %s)", self.email)
        self.model: type[ResourceType] = getattr(self.request.state, 'model')
        self.namespace: str | None = getattr(self.request.state, 'namespace', None)
        self.verb: str = getattr(self.request.state, 'verb')
        self.api_group = self.model.group
        self.plural = self.model.plural
        self.permission = self.get_permission_name(self.api_group, self.plural, self.verb)
        async with self.client as client:
            self.logger.info("Inspecting permissions (server: %s, subject: %s)", client.base_url, self.email)
            self.granted = PermissionSet(await client.permissions(self.namespace, self.permission))

    def is_authenticated(self) -> bool:
        return self.email is not None

    def is_authorized(self) -> bool:
        return self.granted.has(self.permission)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import uuid
from typing import Any

import pydantic

from .resourcename import ResourceName


class ObjectMeta(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    annotations: dict[str, Any] = pydantic.Field(
        default_factory=dict,
        title="Annotations",
        description=(
            "Annotations is an unstructured key value map stored with "
            "a resource that may be set by external tools to store "
            "and retrieve arbitrary metadata. They are not queryable and "
            "should be preserved when modifying objects."
        )
    )
    
    created: datetime.datetime | None = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Created",
        description=(
            "`created` is a timestamp representing the server time "
            "when this object was created. It is not guaranteed to be set in "
            "happens-before order across separate operations. Clients may not "
            "set this value. It is represented in RFC3339 form and is in UTC. "
            "Populated by the system. Read-only. Null for lists."
        ),
        frozen=True
    )
    
    deleted: datetime.datetime | None = pydantic.Field(
        default=None,
        title="Deleted",
        description=(
            "The `deleted` field is an RFC 3339 date and time "
            "at which this resource will be deleted. This field "
            "is set by the server when a graceful deletion is "
            "requested by the user, and is not directly settable "
            "by a client. The resource is expected to be deleted "
            "(no longer visible from resource lists, and not "
            "reachable by name) after the time in this field. "
            "Once set, this value may not be unset or be set "
            "further into the future, although it may be shortened "
            "or the resource may be deleted prior to this time."
        ),
    )
    
    updated: datetime.datetime | None = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Updated",
        description=(
            "`CreationTimestamp` is a timestamp representing the server time "
            "when this object was updated. Clients may not "
            "set this value. It is represented in RFC3339 form and is in UTC. "
            "Populated by the system. Read-only. Null for lists."
        )
    )

    generation: int = pydantic.Field(
        default=1,
        title="Generation",
        description=(
            "A sequence number representing a specific generation "
            "of the state. Populated by the system. Read-only."
        )
    )

    labels: dict[str, str | None] = pydantic.Field(
        default_factory=dict,
        title="Labels",
        description=(
            "Map of string keys and values that can be used to organize and "
            "categorize (scope and select) objects."
        )
    )

    name: ResourceName | str = pydantic.Field(
        default=...,
        title="Name",
        description=(
            "Name must be unique within a namespace. Is required when creating "
            "resources, although some resources may allow a client to request "
            "the generation of an appropriate name automatically. Name is "
            "primarily intended for creation idempotence and configuration "
            "definition. Cannot be updated."
        ),
        frozen=True
    )

    namespace: str = pydantic.Field(
        default='default',
        title="Namespace",
        description=(
            "Namespace defines the space within which each name must "
            "be unique. An empty namespace is equivalent to the `default` "
            "namespace, but `default` is the canonical representation. "
            "Not all objects are required to be scoped to a namespace - "
            "the value of this field for those objects will be empty. "
            "Must be a DNS_LABEL. Cannot be updated."
        )
    )

    tags: set[str] = pydantic.Field(
        default_factory=set
    )

    uid: str = pydantic.Field(
        default_factory=lambda: str(uuid.uuid4()),
        title="UID",
        description=(
            "UID is the unique in time and space value for this object. "
            "It is typically generated by the server on successful creation "
            "of a resource and is not allowed to change on PUT operations. "
            "Populated by the system. Read-only."
        ),
        frozen=True
    )

    @pydantic.field_validator('labels', mode='before')
    def preprocess_labels(cls, values: dict[str, Any] | Any):
        if isinstance(values, dict):
            for label, value in list(values.items()):
                if isinstance(value, str):
                    continue
                if isinstance(value, bool) and value:
                    values[label] = 'true'
                    continue
                if isinstance(value, bool) and not value:
                    values[label] = 'false'
                    continue
                if isinstance(value, int):
                    values[label] = str(value)
                    continue
                if value is None:
                    values[label] = 'null'
                    continue
        return values

    def delete(self) -> None:
        if self.deleted is not None:
            raise ValueError
        self.deleted = datetime.datetime.now(datetime.timezone.utc)
        self.updated = self.deleted

    def merge(self, metadata: 'ObjectMeta') -> None:
        """Merge the annotations, labels and tags of the new
        object into this object.
        """
        self.annotations = {**self.annotations, **metadata.annotations}
        self.labels = {**self.labels, **metadata.labels}
        self.tags.update(metadata.tags)

    def is_labeled(self, names: list[str]) -> bool:
        return all([
            self.labels.get(name) not in {None, 'null'}
            for name in names
        ])

    def set_label(self, name: str, value: Any) -> Any | None:
        if isinstance(value, bool) and value:
            value = 'true'
        if isinstance(value, bool) and not value:
            value = 'false'
        if isinstance(value, int):
            value = str(value)
        if value is None:
            value = 'null'
        if not isinstance(value, str):
            raise TypeError(f'Can not set label of type {type(value).__name__}')
        self.labels[name] = value
        return value


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import Self
from typing import TypeVar

import pydantic

from ..resource import Resource
from ..resource import M


S = TypeVar('S')


class BaseNamespace(Resource[M], Generic[M], kind='Namespace', version='namespaces/v1'):
    __abstract__: ClassVar[bool] = True
    model_config = {'populate_by_name': True}

    type: str = pydantic.Field(
        default=...,
        description=(
            "Used to discriminate between various namespace types."
        )
    )

    @property
    def key(self):
        return self.KeyType(
            name=self.metadata.name,
            namespace=getattr(self.metadata, 'name', None)
        )

    @classmethod
    def __pydantic_init_subclass__(cls, type: str | None = None, **kwargs: Any):
        super().__pydantic_init_subclass__(**kwargs)
        if type is None and not cls.__abstract__:
            raise ValueError(
                f"The `type` class parameter must be specified on {cls.__name__}."
            )
        cls.__abstract__ = False
        cls.model_fields['type'].annotation = Literal[type] # type: ignore
        cls.model_fields['type'].default = type
        cls.model_rebuild()

    def can_change(self, old: Self) -> bool:
        return all([
            old.type == self.type
        ])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical.ext.resource import Resource
from canonical.ext.resource import TransientMeta


class PermissionSet(
    Resource[TransientMeta],
    version='permissionsets.iam.webiam.io/v1'
):
    granted: set[str] = pydantic.Field(
        default_factory=set,
        description=(
            "The set of permissions that are granted on a "
            "resource."
        ),
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Iterable

import fastapi

from canonical.ext.iam import BaseAuthorizationContext
from canonical.ext.iam import ClusterRole
from canonical.ext.iam import ClusterRoleBinding
from canonical.ext.iam import Role
from canonical.ext.iam import RoleBinding
from canonical.ext.iam.types import Permission
from canonical.ext.iam.types import PermissionSet
from canonical.ext.resource import Namespace
from canonical.ext.resource import ResourceType
from .requestemail import RequestEmail
from .resourcerepository import ResourceRepository


class LocalAuthorizationContext(BaseAuthorizationContext):
    logger: logging.Logger = logging.getLogger('uvicorn')
    roles: set[str]
    _cluster_roles: dict[str, ClusterRole] = NotImplemented
    _ready: bool = False

    def __init__(
        self,
        repo: ResourceRepository,
        request: fastapi.Request,
        email: RequestEmail,
    ):
        self._cluster_permissions = LocalAuthorizationContext._cluster_roles
        self.email = email
        self.granted = PermissionSet()
        self.repo = repo
        self.request = request
        self.subject_type = 'User'

    def get_cluster_bindings(self):
        self.logger.debug(
            "Retrieving global role bindings (kind: %s, email: %s)",
            self.subject_type,
            self.email
        )
        return self.repo.query(
            model=ClusterRoleBinding,
            filters=[
                ('subjects.kind', '=', self.subject_type),
                ('subjects.name', '=', str(self.email)),
            ]
        )

    def get_namespace_bindings(self):
        self.logger.debug(
            "Retrieving local role bindings (namespace: %s, kind: %s, email: %s)",
            self.namespace,
            self.subject_type,
            self.email
        )
        return self.repo.query(
            model=RoleBinding,
            filters=[
                ('subjects.kind', '=', self.subject_type),
                ('subjects.name', '=', str(self.email)),
            ],
            namespace=self.namespace
        )

    def get_namespace_roles(self, roles: Iterable[str]):
        assert self.namespace is not None
        return self.repo.query(
            model=Role,
            # TODO: Current implementation causes the query
            # to return not any object if there are multiple
            # roles.
            #filters=[('metadata.name', '=', list(roles))],
            namespace=self.namespace
        )

    def is_authenticated(self) -> bool:
        return self.email is not None

    def is_authorized(self) -> bool:
        return self.is_granted(self.permission)

    def is_granted(self, permission: str) -> bool:
        return self.granted.has(permission)

    async def get_permissions(self, permissions: set[str]) -> PermissionSet:
        return PermissionSet({Permission(p) for p in permissions if self.is_granted(p)})

    async def get_namespace_permissions(self, roles: Iterable[str]) -> PermissionSet:
        permissions = PermissionSet()
        async for role in self.get_namespace_roles(roles):
            if role.metadata.name not in roles:
                continue
            permissions.update(role.permissions)
        return permissions

    async def has(self, permissions: str | set[str]) -> bool:
        if isinstance(permissions, str):
            permissions = {permissions}
        granted = await self.get_permissions(permissions)
        return set(granted) == permissions

    async def setup(self):
        self.model: type[ResourceType] = getattr(self.request.state, 'model')
        self.namespace: str | None = getattr(self.request.state, 'namespace', None)
        self.verb: str = getattr(self.request.state, 'verb')
        self.api_group = self.model.group
        self.plural = self.model.plural
        self.permission = self.get_permission_name(self.api_group, self.plural, self.verb)

        # Namespace is a special cause because permissions in the namespace
        # also apply to the namespace itself.
        if self.namespace is None and self.model == Namespace:
            # This will be none for cluster resources.
            self.namespace = getattr(self.request.state, 'name', None)
    
        await self.setup_cluster()

        scoped_roles: set[str] = set()
        global_roles: set[str] = set()
        async for binding in self.get_cluster_bindings():
            role = self._cluster_roles[binding.role_ref.name]
            self.granted.update(role.permissions)
        if self.namespace is not None:
            async for obj in self.get_namespace_bindings():
                if obj.is_global():
                    global_roles.add(obj.role_ref.name)
                else:
                    scoped_roles.add(obj.role_ref.name)

            self.granted |= await self.get_namespace_permissions(scoped_roles)

    async def setup_cluster(self):
        if self._cluster_roles != NotImplemented:
            return
        self.logger.debug("Retrieving cluster roles and permissions")
        self._cluster_roles = {}
        async for role in self.repo.all(ClusterRole):
            self._cluster_roles[role.metadata.name] = role


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

from canonical.ext.resource import NamespacedObjectMeta
from .baserolebinding import BaseRoleBinding


class RoleBinding(
    BaseRoleBinding[NamespacedObjectMeta[str]],
    kind=Literal['ClusterRole', 'Role'],
    version='rolebindings.iam.webiam.io/v1'
):
    pass


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


__all__: list[str] = [
    'ResourceOptions'
]


class ResourceOptions(pydantic.BaseModel):
    methods: list[str] = pydantic.Field(
        default_factory=list,
        description=(
            "The list of allowed HTTP methods that may be invoked to "
            "perform operations related to the resource."
        )
    )

    permissions: list[str] = pydantic.Field(
        default_factory=list,
        description=(
            "If the request is authenticated, the list of permissions "
            "that the authenticated subject has on the resource."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical.ext.resource import ResourceType
from .create import Create
from ..params import ResourceRepository


class DefaultCreate(Create[ResourceType]):
    creates = True
    detail = False
    exists = False
    method = 'POST'
    requires_body = True
    status_code = 201
    verb = 'create'

    async def handle(
        self,
        repo: ResourceRepository,
        obj: ResourceType
    ) -> ResourceType:
        return await obj.persist(repo)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import inspect
from typing import cast
from typing import get_args
from typing import get_origin
from typing import Any
from typing import Callable
from typing import Generic
from typing import TypeVar
from typing import Union
from typing import Unpack
from typing import TYPE_CHECKING

import fastapi
import fastapi.params
from fastapi.exceptions import HTTPException

from canonical.ext.jsonpatch import JSONPatchType
from canonical.ext.iam import PermissionQuery
from canonical.ext.iam import PermissionSet
from canonical.ext.resource import Error
from canonical.ext.resource import Resource
from canonical.ext.resource import RootResource
from canonical.utils import merge_signatures
from .apiroute import APIRoute
from .params import AcceptedContentType
from .params import NegotiateResponseMediaType
from .params import RequestAuthorizationContext
from .params import RequestResource
from .params import RequestVerb
from .params import ResourceRepository
from .resourceoptions import ResourceOptions
from .response import Response
from . import verbs
if TYPE_CHECKING:
    from .resourceapplication import ResourceApplication


R = TypeVar('R', bound=Resource[Any] | RootResource[Any])
ResponseModel = TypeVar('ResponseModel')




class ResourceRouterMetaclass(type):

    def __getitem__(cls, model: type[R]) -> type['ResourceRouter[R]']:
        return type(f'{model.__name__}', (cls,), { # type: ignore
            'model': model
        })


class ResourceRouter(fastapi.APIRouter, Generic[R]):
    detail_verbs: set[str] = {
        'authorize',
        'destroy',
        'exists',
        'get',
        'inspect',
        'replace',
        'update',
    }
    response_class: type[Response[R]]
    resource_model: type[R]

    @classmethod
    def new(cls, model: type[R], **kwargs: Any) -> 'ResourceRouter[R]':
        return cls.typed(model)(**kwargs)

    @classmethod
    def typed(cls, model: type[R]) -> type['ResourceRouter[R]']:
        impl: type[ResourceRouter[R]]  = type(f'{model.__name__}Router', (cls,), {
            'resource_model': model
        })
        return impl

    @classmethod
    def with_model(cls, app: ResourceApplication, model: type[R], **kwargs: Any) -> 'ResourceRouter[R]':
        models: tuple[type[Resource[Any]] | type[RootResource[Any]], ...] = tuple([])
        if get_origin(model) == Union:
            models = get_args(model)
        if not models:
            models = tuple([model])
        router = fastapi.APIRouter()
        for m in models: # type: ignore
            impl = type(f'{m.__name__}Router', (cls,), {
                'resource_model': m
            })
            child = impl()
            child.add_to_application(app, **kwargs)
            router.include_router(router=child)
        return cast(ResourceRouter[model], router)

    def __init__(
        self,
        enabled_verbs: set[str] | None = None,
        granter: bool = False,
        **kwargs: Any
    ):
        self.verbs_enabled = set(enabled_verbs or [])
        deps: list[Any] = kwargs.setdefault('dependencies', [])
        deps.extend([
            NegotiateResponseMediaType({
                'text/html',
                'application/yaml',
                'application/json'
            }),
            *self.inject_defaults()
        ])
        self.collect_dependencies(deps, None)
        model = self.resource_model

        super().__init__(route_class=APIRoute, **kwargs)
        self.response_class = Response.typed(model)
        self.resource_model = model
        self.verbs: list[verbs.Verb[Any]] = []

        self.verbs.append(verbs.Exists(self.resource_model))
        if 'create' in self.verbs_enabled:
            self.verbs.append(verbs.DefaultCreate(self.resource_model))

        if granter:
            self.add_resource_route(
                method='POST',
                verb='authorize',
                detail=True,
                path=f'/{model.base_path}/{{name}}:permissions',
                endpoint=self.permissions,
                summary=f'Test {model.__name__} permissions',
                description=(
                    f"Get the permissions granted to the authenticated "
                    f"subject on a specific `{model.__name__}` object."
                ),
                response_model=PermissionSet,
                response_description=f"PermissionSet object.",
                responses={
                    401: {
                        'model': Error,
                        'description': (
                            "The provided request credential is expired, not effective"
                            ", or otherwise malformed."
                        )
                    },
                    403: {
                        'model': Error,
                        'description': "Untrusted credential."
                    },
                },
                authenticated=False,
                with_resource=False
            )
        return
        self.add_resource_route(
            method='GET',
            verb='get',
            detail=True,
            path=f'/{model.base_path}/{{name}}',
            endpoint=self.retrieve,
            summary=f'Retrieve a specific {model.__name__}',
            description=f"Retrieve a `{model.__name__}` object.",
            response_model=model,
            response_description=f"{model.__name__} object.",
        )



        self.add_api_route(
            methods=['PATCH'],
            path=f'/{model.base_path}/{{name}}',
            endpoint=self.update,
            summary=f'Update endpoint',
            description=f"Update a `{model.__name__}` object.",
            dependencies=[
                AcceptedContentType({'application/yaml', 'application/json'}),
                fastapi.Depends(self.set_verb('update'))
            ],
            response_model=model,
            status_code=205,
            responses={
                409: {
                    'model': Error,
                    'description': (
                        f'One or more of the patches can not be applied to the {model.__name__}.'
                    )
                }
            },
            response_description=f"Updated {model.__name__} object.",
            response_model_by_alias=True,
            tags=[model.__name__]
        )

        if model.is_destroyable():
            self.add_api_route(
                methods=['DELETE'],
                path=f'/{model.base_path}/{{name}}',
                endpoint=self.destroy,
                summary=f'Destroy endpoint',
                description=f"Destroy a `{model.__name__}` object.",
                dependencies=[
                    fastapi.Depends(self.set_verb('destroy'))
                ],
                response_model=model,
                status_code=200,
                responses={
                    409: {
                        'model': Error,
                        'description': (
                            f'The {model.__name__} identified by the path parameters '
                            'can not be destroyed.'
                        )
                    }
                },
                response_description=f"Last version of the {model.__name__} object.",
                response_model_by_alias=True,
                tags=[model.__name__]
            )

        self.add_resource_route(
            method='GET',
            verb='list',
            path=f'/{model.base_path}',
            endpoint=self.collection,
            summary=f'Retrieve a list of {model.__name__} objects',
            description=f"Retrieve a list of `{self.resource_model.List.__name__}` objects.",
            response_model=model.List,
            response_description=f"{model.List.__name__} object.",
        )
        if model.is_purgable():
            self.add_api_route(
                methods=['DELETE'],
                path=f'/{model.base_path}',
                endpoint=self.purge,
                summary=f'Destroy every {model.__name__}',
                description=f"Purge a collection of `{model.__name__}` objects.",
                dependencies=[
                    fastapi.Depends(self.set_verb('destroy'))
                ],
                response_model=model,
                status_code=202,
                responses={
                    409: {
                        'model': Error,
                        'description': (
                            f'The {model.List.__name__} identified by the path parameters '
                            'can not be purged.'
                        )
                    }
                },
                response_description=f"No content.",
                response_model_by_alias=True,
                tags=[model.__name__]
            )

        self.add_resource_route(
            method='OPTIONS',
            verb='inspect',
            detail=True,
            path=f'/{model.base_path}/{{name}}',
            endpoint=self.inspect,
            summary=f'{model.__name__} metadata',
            description=f"Retrieve metadata describing the resource type.",
            responses={
                200: {
                    'model': ResourceOptions,
                    'headers': {
                        'Accept': {
                            'description': (
                                "A comma-separated list describing the allowed "
                                "HTTP methods on this endpoint."
                            )
                        }
                    }
                },
                401: {
                    'model': Error,
                    'description': (
                        "The provided request credential is expired, not effective"
                        ", or otherwise malformed."
                    )
                },
                403: {
                    'model': Error,
                    'description': "Untrusted credential."
                },
            },
            response_description=f"ResourceOptions object.",
            response_model=ResourceOptions,
        )


    def add_to_application(self, app: ResourceApplication):
        if not app.openapi_tags:
            app.openapi_tags = []
        for verb in self.verbs:
            verb.add_to_router(self, app.register_resource)
        app.openapi_tags.append({
            'name': self.resource_model.__name__,
            'description': getattr(self.resource_model, '__doc__', None)
        })
        app.include_router(router=self)

    def add_resource_route(
        self,
        *,
        verb: str,
        method: str,
        path: str,
        summary: str,
        description: str,
        endpoint: Any,
        response_model: type[Any] | None = None,
        response_description: str,
        responses: dict[int, Any] | None = None,
        dependencies: list[Any] | None = None,
        detail: bool = False,
        authenticated: bool = True,
        with_resource: bool = True
    ) -> None:
        model = self.resource_model
        responses = dict(responses or {})
        dependencies = dependencies or []
        if verb in self.detail_verbs:
            detail = True
        if verb in {'create', 'replace', 'update', 'authorize'}:
            dependencies.append(AcceptedContentType({'application/yaml', 'application/json'}))
            responses.update({
                415: {
                    'model': Error,
                    'description': "Invalid content type for request body."
                },
            })
        if detail:
            responses.update({
                404: {
                    'model': Error,
                    'description': (
                        f'The {model.__name__} specified by the path parameter(s) '
                        'does not exist.'
                    )
                }
            })
            if with_resource:
                dependencies.insert(0, self.inject_resource())
        if not authenticated:
            responses.pop(401, None)
        self.add_api_route(
            methods=[method],
            path=path,
            endpoint=endpoint,
            summary=summary,
            description=description,
            responses={**responses},
            response_model=self.resource_model or response_model,
            response_model_by_alias=True,
            response_description=response_description,
            dependencies=[
                fastapi.Depends(self.set_verb(verb)),
                *dependencies,
                fastapi.Depends(self.authorize),
            ],
            tags=[model.__name__],
        )
        #self.register_resource(path, self.resource_model)

    def add_verb(
        self,
        cls: type[verbs.Verb[Any]],
        permissions: set[str] | None = None
    ) -> None:
        self.verb_factory(cls, permissions=permissions)

    def collect_dependencies(self, dependencies: list[Any], verb: str | None):
        pass

    def get_media_type(self, request: fastapi.Request) -> str:
        return getattr(request.state, 'media_type')

    def inject_defaults(self) -> list[fastapi.params.Depends]:
        return [
            fastapi.params.Depends(self.inject_model),
            fastapi.params.Depends(self.inject_namespace),
            fastapi.params.Depends(self.inject_resource),
        ]

    def inject_model(self, request: fastapi.Request):
        setattr(request.state, 'model', self.resource_model)

    def inject_namespace(self, request: fastapi.Request):
        if 'name' in request.path_params:
            setattr(request.state, 'name', request.path_params['name'])
        if 'namespace' in request.path_params:
            setattr(request.state, 'namespace', request.path_params['namespace'])

    def inject_resource(self):
        return self._inject_resource_namespaced()\
            if self.resource_model.is_namespaced()\
            else self._inject_resource_cluster()

    def writer_factory(self, model: type[R], func: Any, partial: bool = False):
        async def f(resource: model, *args: Any, **kwargs: Any) -> Any:
            return await func(resource, *args, **kwargs)
        f.__signature__ = merge_signatures([ # type: ignore
            inspect.signature(func),
            inspect.signature(f)
        ])
        return f

    def set_verb(self, verb: str):
        return self.inject_verb(verb)

    def inject_verb(self, verb: str):
        def f(request: fastapi.Request):
            setattr(request.state, 'verb', verb)
        return f

    def subresource(self, methods: list[str], path: str, verb: str, **kwargs: Any):
        assert not str.startswith(path, '/')
        path = f'/{self.resource_model.base_path}/{{name}}/{path}'
        dependencies: list[Any] = kwargs.setdefault('dependencies', [])
        dependencies.insert(0, fastapi.Depends(self.inject_resource))
        dependencies.insert(0, fastapi.Depends(self.authorize))
        dependencies.insert(0, fastapi.Depends(self.inject_verb(verb)))
        return self.api_route(
            path=path,
            methods=methods,
            tags=[self.resource_model.__name__],
            **kwargs
        )
        
    async def authorize(
        self,
        ctx: RequestAuthorizationContext,
        verb: RequestVerb
    ) -> None:
        if verb == 'authorize':
            return
        if not ctx.is_authenticated():
            raise HTTPException(
                status_code=401,
                detail="Authentication required."
            )
        if not ctx.is_authorized():
            raise HTTPException(
                status_code=403,
                detail=(
                    "The request subject is not granted permission "
                    "to perform this operation."
                )
            )

    async def collection(self) -> Resource[Any]:
        raise NotImplementedError

    async def create(self, resource: R) -> Resource[Any]:
        raise NotImplementedError

    async def destroy(self) -> Resource[Any]:
        raise NotImplementedError

    async def permissions(
        self,
        request: fastapi.Request,
        ctx: RequestAuthorizationContext,
        query: PermissionQuery
    ) -> Response[Any]:
        return self.render_to_response(
            request=request,
            status_code=200,
            instance=PermissionSet.model_validate({
                'granted': await ctx.get_permissions({str(p) for p in query.spec.permissions})
            })
        )

    async def purge(self) -> Resource[Any]:
        raise NotImplementedError

    async def replace(self, resource: R) -> Resource[Any]:
        raise NotImplementedError

    async def retrieve(
        self,
        request: fastapi.Request,
        resource: RequestResource,
    ) -> Response[Any]:
        return self.render_to_response(
            request=request,
            status_code=200,
            instance=resource
        )

    def render_to_response(self, *, request: fastapi.Request, instance: Any, status_code: int = 200) -> Response[Any]:
        return Response(
            status_code=status_code,
            media_type=self.get_media_type(request),
            content=instance
        )

    def verb(
        self,
        cls: type[verbs.Verb[Any]],
        *,
        verb: str | None = None,
        **kwargs: Unpack[verbs.VerbParameters]
    ) -> Callable[..., Any]:
        def f(func: Callable[..., Any]) -> Callable[..., Any]:
            impl: verbs.Verb[[], Any] = type(cls.__name__, (cls,), { # type: ignore
                'handle': func,
                'verb': verb or getattr(cls, 'verb')
            })
            self.verb_factory(impl, **kwargs) # type: ignore
            return func
        return f

    def verb_factory(self, cls: type[verbs.Verb[Any]], *args: Any, **kwargs: Any):
        if any([v.verb == cls.verb for v in self.verbs]):
            raise ValueError(f"Verb '{cls.verb}' is already declared.")
        self.verbs.append(cls(model=self.resource_model, *args, **kwargs)) # type: ignore
        self.verbs_enabled.add(cls.verb)

    async def update(
        self,
        patch: JSONPatchType,
        name: str = fastapi.Path(...)
    ) -> Resource[Any]:
        raise NotImplementedError

    async def inspect(self) -> ResourceOptions:
        raise NotImplementedError

    def _inject_resource_cluster(self) -> Any:
        async def f(
            request: fastapi.Request,
            repo: ResourceRepository,
            name: str = fastapi.Path(
                description=f'The `.metadata.name` of an existing {self.resource_model.__name__}.',
                max_length=64,
            )
        ):
            try:
                resource = await repo.get_by_name(self.resource_model, name)
                setattr(request.state, 'resource', resource)
            except repo.DoesNotExist:
                raise HTTPException(
                    status_code=404,
                    detail=(
                        f"The {self.resource_model.__name__} specified by "
                        f"the path parameters does not exist: '{name}'."
                    )
                )

        return fastapi.Depends(f)

    def _inject_resource_namespaced(self) -> Any:
        async def f(
            repo: ResourceRepository,
            request: fastapi.Request,
            namespace: str = fastapi.Path(
                description=(
                    "Identifies the namespace that contains the "
                    f"`{self.resource_model.__name__}`."
                )
            ),
            name: str = fastapi.Path(
                description=(
                    f'The `.metadata.name` of an existing '
                    f'`{self.resource_model.__name__}`.'
                )
            )
        ):
            try:
                resource = await repo.get_by_name(self.resource_model, name, namespace=namespace)
                setattr(request.state, 'resource', resource)
            except repo.DoesNotExist:
                raise HTTPException(
                    status_code=404,
                    detail=(
                        f"The {self.resource_model.__name__} specified by "
                        f"the path parameters does not exist: '{name}'."
                    )
                )

        return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from collections import OrderedDict
from inspect import Parameter
from inspect import Signature
from typing import Any
from typing import TypeVar

from canonical.ext.resource import Error
from canonical.ext.resource import Resource
from canonical.ext.resource import RootResource
from ..response import Response
from .default import Default


T = TypeVar('T', bound=Resource[Any] | RootResource[Any])


class Create(Default[T]):
    creates = True
    detail = False
    exists = False
    method = 'POST'
    requires_body = True
    status_code = 201
    verb = 'create'

    def annotate_handler(self, signature: Signature) -> Signature:
        params: OrderedDict[str, Parameter] = OrderedDict()
        for name, param in signature.parameters.items():
            if param.annotation != T:
                params[name] = param
                continue
            params[name] = param.replace(annotation=self.model)

        return signature.replace(parameters=list(params.values()))

    def get_endpoint_summary(self) -> str:
        return f'Create a new {self.model.__name__}'

    def get_openapi_responses(self) -> dict[int, dict[str, Any]]:
        return {
            **super().get_openapi_responses(),
            409: {
                'model': Error,
                'description': (
                    f'Conflicts with an existing {self.model.__name__} object.'
                )
            }
        }

    def get_response_description(self) -> str:
        return f'The created {self.model.__name__} object.'

    async def render_to_response(
        self,
        result: Any
    ) -> Response[Any]:
        if not isinstance(result, (Resource, RootResource)):
            raise TypeError(
                "Function handle() must return a Resource or RootResource "
                "instance."
            )
        return Response(
            status_code=201,
            content=result
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Iterable

import fastapi
import fastapi.params
from fastapi.exceptions import HTTPException

from canonical.utils.http import MediaTypeSelector


def AcceptedContentType(available: Iterable[str]) -> fastapi.params.Depends:
    selector = MediaTypeSelector(available)

    def f(request: fastapi.Request):
        header = request.headers.get('Content-Type')
        if header is None:
            raise HTTPException(
                status_code=415,
                detail="The Content-Type header is required."
            )
        selected = selector.select(header)
        if selected is None:
            raise HTTPException(
                status_code=415,
                detail=(
                    "The server refuses to accept the request because "
                    "the payload format is in an unsupported format."
                )
            )
        return selected

    return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import cast
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic

from .domainname import DomainName
from .stringtype import StringType


T = TypeVar('T', bound='ResourceName')
M = TypeVar('M', bound=pydantic.BaseModel)


class ResourceName(StringType):
    model: Any
    relname: str
    service: DomainName
    patterns = [re.compile(r'//.*')]

    @property
    def id(self) -> str:
        return str.split(self, '/')[-1]

    @property
    def kind(self) -> str:
        return str.split(self.relname, '/')[0]

    @classmethod
    def null(cls: type[T]) -> T: # pragma: no cover
        """Return an instance that represents an unassigned
        resource name.
        """
        return cls('//cochise.io/_/null')

    @classmethod
    def subclass(cls, model: type[Any]):
        return type(f'{model.__name__}Name', (cls,), {}, model=model)

    def __new__(
        cls: type[T],
        object: str,
        service: DomainName | None = None,
        relname: str | None = None
    ) -> T:
        self = super().__new__(cls, object) # type: ignore
        if service is None:
            # Was not created by pydantic
            for validator in cls.__get_validators__():
                self = validator(self)
        else:
            assert relname is not None
            self.relname = relname
            self.service = service
        return cast(T, self)

    def __init_subclass__(cls, model: type[Any]) -> None:
        super().__init_subclass__()
        cls.model = model

    @classmethod
    def validate(cls, v: str, _: Any = None) -> str:
        v = super().validate(v, _)
        assert str.startswith(v, '//')
        service, _, relname = str.partition(v[2:], '/')
        if not relname:
            raise ValueError("a valid ResourceName contains a relative name.")
        for validator in DomainName.__get_validators__():
            service = validator(service)
        return cls(v, relname=relname, service=DomainName(service))


class TypedResourceName(ResourceName, Generic[M], model=pydantic.BaseModel):
    model: type[M]


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import decimal
from typing import Any

import pydantic
from canonical import ResourceName
from canonical import Text


class SKUDescriptor(pydantic.BaseModel):
    model_config = {'populate_by_name': True}
    _resources: dict[str, ResourceName] = pydantic.PrivateAttr()

    code: str = pydantic.Field(
        default='',
        description=(
            "The code representing the stock keeping unit."
        ),
    )

    country_of_origin: str | None = pydantic.Field(
        default=None,
        alias='countryOfOrigin',
        title='Country of Origin',
        description=(
            "Country of origin of the product, if applicable."
        ),
        max_length=2
    )

    hs: str | None = pydantic.Field(
        default=None,
        title='Harmonized System (HS)',
        description=(
            "Harmonized System (HS) code of the product.\n\n"
            "The Harmonized Commodity Description and Coding System, "
            "also known as the Harmonized System (HS) of tariff "
            "nomenclature is an internationally standardized system of "
            "names and numbers to classify traded products. It came into "
            "effect in 1988 and has since been developed and maintained "
            "by the World Customs Organization (WCO) (formerly the "
            "Customs Co-operation Council), an independent "
            "intergovernmental organization based in Brussels, Belgium."
        )
    )

    product_name: Text = pydantic.Field(
        default=...,
        alias='productName',
        title='Product name',
        description=(
            "The internal display name of the product or service. May differ "
            "from what is shown to customers."
        )
    )

    ean: str | None = pydantic.Field(
        default=None,
        title='EAN',
        description=(
            "The International Article Number, (also known as European "
            "Article Number or EAN) is a standard describing a barcode "
            "symbology and numbering system used in global trade to "
            "identify a specific retail product type, in a specific "
            "packaging configuration, from a specific manufacturer. "
        ),
        max_length=13
    )

    external_resources: list[ResourceName] = pydantic.Field(
        default_factory=list,
        alias='externalResources',
        title='External references',
        description=(
            "Lists the external resources that depend on this SKU."
        )
    )

    weight: decimal.Decimal | None = pydantic.Field(
        default=None,
        title='Weight',
        description="Weight of the product in grams."
    )

    @property
    def pk(self) -> str:
        return self.code

    @property
    def resources(self) -> dict[str, ResourceName]:
        return self._resources

    def add_resource(self, name: str | ResourceName) -> None:
        if not isinstance(name, ResourceName):
            name = ResourceName(name)
        if name in self.external_resources:
            return
        self.external_resources.append(name)
        self._resources[name.service] = name

    def model_post_init(self, _: Any) -> None:
        self._resources = {x.service: x for x in self.external_resources}

    def has(self, service: str, kind: str | None = None) -> bool:
        if kind is None:
            result = service in self._resources
        else:
            result = any([
                x.service == service and x.kind == kind
                for x in self.external_resources
            ])
        return result


# Copyright 2019 Tim Heap
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following
# conditions:
# 
# The above copyright notice and this permission notice shall
# be included in all copies or substantial portions of the
# Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import re
from decimal import Decimal
from itertools import product
from typing import Any
from typing import Iterable


# Matches 'text' or 'application'
major_type_str = r'[a-zA-Z0-9._-]+'

# Matches 'html', or 'xml+gml'
minor_type_str = r'[a-zA-Z0-9._+-]+'

# Matches either '*', 'image/*', or 'image/png'
valid_mime_type = re.compile(r'^(?:\*|{major_type}/\*|{major_type}/{minor_type})$'.format(
    major_type=major_type_str, minor_type=minor_type_str))

# Matches the 'q=1.23' from the parameters of a Accept mime types
q_match = re.compile(r'(?:^|;)\s*q=([0-9.-]+)(?:$|;)')


class AcceptableType:
    mime_type: str
    pattern: re.Pattern[str]
    weight = Decimal(1)

    def __init__(self, raw_mime_type: str):
        bits = raw_mime_type.split(';', 1)

        mime_type = bits[0]
        if not valid_mime_type.match(mime_type):
            raise ValueError('"%s" is not a valid mime type' % mime_type)

        tail = ''
        if (len(bits) > 1):
            tail = bits[1]

        self.mime_type = mime_type
        self.weight = get_weight(tail)
        self.pattern = get_pattern(mime_type)

    def matches(self, mime_type: str):
        return self.pattern.match(mime_type)

    def __str__(self):
        display = self.mime_type
        if self.weight != Decimal(1):
            display += '; q=%0.2f' % self.weight

        return display

    def __repr__(self):
        return '<AcceptableType {0}>'.format(self)

    def __eq__(self, other: Any):
        if not isinstance(other, AcceptableType):
            return NotImplemented
        return (self.mime_type, self.weight) == (other.mime_type, other.weight)

    def __lt__(self, other: Any):
        if not isinstance(other, AcceptableType):
            return NotImplemented
        return self.weight < other.weight


def get_best_match(header: str, types: Iterable[str]):
    """Find the best mime type to respond to a request with,
    from an ``Accept`` header and list of response mime types
    the application supports.

    .. code-block:: python

        return_data = reticulate_splines()
        return_type = get_best_match(
            request.META.get('HTTP_ACCEPT'),
            ['text/html', 'application/xml', 'text/json'])
        if return_type == 'text/html':
            return render_template('response.html', return_data)
        if return_type == 'application/xml':
            return convert_to_xml(return_data)
        if return_type == 'application/json':
            return json.dumps(return_data)

    :param header: The HTTP ``Accept`` header from the request.
    :type header: str
    :param available_types: The response mime types the application supports.
    :type available_types: List[str]
    :return: The best matching mime type from the supported response
        mime types, or ``None`` if there are no valid matches.
    """
    accepted = parse_header(header)

    for acceptable, available in product(accepted, types):
        if acceptable.matches(available):
            return available

    return None


def parse_header(header: str):
    """Parse an ``Accept`` header into a sorted list of :class:`AcceptableType`
    instances.
    """
    raw_mime_types = header.split(',')
    mime_types: list[AcceptableType] = []
    for raw_mime_type in raw_mime_types:
        try:
            mime_types.append(AcceptableType(raw_mime_type.strip()))
        except ValueError:
            pass

    return sorted(mime_types, reverse=True)


def get_weight(tail: str):
    """Given the tail of a mime type header (the bit after the first ``;``),
    find the ``q`` (weight, or quality) parameter.

    If no valid ``q`` parameter is found, default to ``1``, as per the spec.
    """
    match = re.search(q_match, tail)
    if match:
        try:
            return Decimal(match.group(1))
        except ValueError:
            pass

    # Default weight is 1
    return Decimal(1)


def get_pattern(mime_type: str) -> re.Pattern[str]:
    """Convert a mime type string into a regexp pattern that matches that mime
    type. Supports both literal mime types such as ``image/png``, and wildcard
    mime types such as ``image/*``.

    :param mime_type: The mime type string to build a pattern for
    :type mime_type: str
    :return: A regexp that matches the mime_type string.
    """
    # *: Simple match all case
    if mime_type in {'*', '*/*'}:
        return valid_mime_type
    # image/*: Match the major type
    if mime_type.endswith('*'):
        return re.compile('^' + re.escape(mime_type[:-1]) + minor_type_str + '$')
    # All other cases, match the exact mime type string
    return re.compile('^' + re.escape(mime_type) + '$')



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

import fastapi

from ..params import ResourceRepository
from ..params import RequestAuthorizationContext
from ..params import RequestResourceName
from .default import Default


T = TypeVar('T')


class Exists(Default[T]):
    detail = True
    existing = False
    method = 'HEAD'
    status_code = 200
    verb = 'exists'

    async def authorize(self) -> None: # type: ignore
        pass

    async def exists(
        self,
        repo: ResourceRepository,
        key: RequestResourceName
    ) -> bool:
        self.logger.info(
            "Inspecting %s existence (key: %s)",
            self.model.__name__, key
        )
        return await repo.exists(key)

    async def handle( # type: ignore
        self,
        ctx: RequestAuthorizationContext,
        repo: ResourceRepository,
        key: RequestResourceName,
    ) -> bool:
        if not ctx.is_authenticated() or not ctx.is_authorized():
            return False
        return await self.exists(repo, key)

    async def render_to_response(self, result: bool) -> fastapi.Response: # type: ignore
        return fastapi.Response(
            status_code=200 if result else 404
        )

    def get_endpoint_summary(self) -> str:
        return f'Check if {self.model.__name__} exists'

    def get_openapi_responses(self) -> dict[int, dict[str, Any]]:
        return {
            **super().get_openapi_responses(),
            200: {
                'description': self.get_response_description()
            }
        }

    def get_response_description(self) -> str:
        return f'The {self.model.__name__} specified by the path parameter(s) exists.'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import Protocol
from typing import TypeVar


T = TypeVar('T')


class ITyped(Protocol, Generic[T]):
    __module__: str = 'canonical.protocols'
    __qualname__: str = f'{__module__}.ITyped'

    @classmethod
    def typed(cls, t: type[Any]) -> type['ITyped[T]']:
        ...

    def get_type(self) -> type[T]:
        ...


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

from canonical.security import DomainACL


class Model(pydantic.BaseModel):
    acl: DomainACL


@pytest.mark.parametrize("values", [
    ([]),
    (["example.com"]),
    (["~example.com"]),
    (["example.com", "~example.com"]),
])
def test_bool_true(values: list[str]):
    acl = DomainACL.fromiterable(values)
    assert acl


def test_bool_false():
    acl = DomainACL.null()
    assert not acl


def test_empty_does_allow_any():
    acl = DomainACL.null()
    assert not acl
    assert acl.allows("example.com")
    

@pytest.mark.parametrize("domain,allows", [
    ("example.com", True),
    ("example.net", False),
])
def test_allows(domain: str, allows: bool):
    acl = DomainACL.fromiterable([
        "example.com",
        "~example.net"
    ])
    assert acl.allows(domain) == allows
    

@pytest.mark.parametrize("domain,allows", [
    ("example.com", True),
    ("google.com", True),
    ("microsoft.com", True),
    ("gmail.com", True),
    ("outlook.com", True),
    ("example.net", False),
])
def test_allows_all_except_denied(domain: str, allows: bool):
    acl = DomainACL.fromiterable([
        "ALL",
        "~example.net"
    ])
    assert acl.allows(domain) == allows


def test_serialize():
    acl = DomainACL.fromiterable([
        "example.com",
        "~example.net"
    ])
    assert set(acl.serialize()) == {"example.com", "~example.net"} # type: ignore


def test_serialize_null():
    acl = DomainACL.null()
    assert acl.serialize() is None


@pytest.mark.parametrize("value", [
    ([]),
    (["example.com"]),
    None
])
def test_model_validate(value: list[str] | None):
    obj = Model.model_validate({
        'acl': value
    })
    assert isinstance(obj.acl, DomainACL)


@pytest.mark.parametrize("value", [
    ([]),
    (["example.com"]),
    None
])
def test_model_validate_json(value: list[str] | None):
    obj1 = Model(acl=value).model_dump_json() # type: ignore
    obj2 = Model.model_validate_json(obj1)
    assert obj2.acl.serialize() == value


@pytest.mark.parametrize("value", [
    ([]),
    (["example.com"]),
    None
])
def test_model_init(value: list[str] | None):
    obj = Model(acl=value) # type: ignore
    assert isinstance(obj.acl, DomainACL)


@pytest.mark.parametrize("value", [
    ([]),
    (["example.com"]),
    None
])
def test_model_dump(value: list[str] | None):
    obj = Model(acl=value) # type: ignore
    assert obj.model_dump()['acl'] == value


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import NotRequired
from typing import TypedDict

from .base import BaseOperation as Verb
from .create import Create
from .defaultcreate import DefaultCreate
from .exists import Exists
from .get import Get
from .replace import Replace
from .retrieve import Retrieve


__all__: list[str] = [
    'Create',
    'DefaultCreate',
    'Exists',
    'Get',
    'Replace',
    'Retrieve',
    'Verb'
]


class VerbParameters(TypedDict):
    description: NotRequired[str]
    method: NotRequired[str]
    path: NotRequired[str]
    permissions: NotRequired[set[str]]
    summary: NotRequired[str]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from collections import OrderedDict
from typing import get_origin
from typing import Any
from typing import Annotated

import fastapi
import fastapi.params

from canonical.ext.resource import IResourceRepository
from .resourceapplication import ResourceApplication
from .resourceclient import ResourceClient
from .resourcerouter import ResourceRouter


__all__: list[str] = [
    'ResourceApplication',
    'ResourceClient',
    'ResourceRouter'
]

def setup_dependencies(
    repository: type[IResourceRepository]
):
    async def f(
        request: fastapi.Request,
        resources: IResourceRepository = fastapi.Depends(repository)
    ):
        setattr(request.state, 'resources', resources)

    return [fastapi.Depends(f)]


def inject(*args: Any, **kwargs: Any):
    def f(request: fastapi.Request, **kwargs: Any):
        for k, v in kwargs.items():
            setattr(request.state, k, v)

    sig = inspect.signature(f)
    params: OrderedDict[str, inspect.Parameter] = OrderedDict([
        ('request', sig.parameters['request'])
    ])
    for i, dependency in enumerate(args):
        name = f'__arg_{i}'
        if isinstance(dependency, fastapi.params.Depends):
            kwargs[name] = dependency
            continue
        if not get_origin(dependency) == Annotated:
            continue
        params[name] = inspect.Parameter(
            kind=inspect.Parameter.POSITIONAL_ONLY,
            name=name,
            annotation=dependency
        )
        
    for name, dependency in kwargs.items():
        params[name] = inspect.Parameter(
            kind=inspect.Parameter.KEYWORD_ONLY,
            name=name,
            default=dependency,
            annotation=Any
        )
    setattr(f, '__signature__', sig.replace(parameters=list(params.values())))
    return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

from .default import Default


T = TypeVar('T')


class Get(Default[T]):
    creates = False
    existing = False
    detail = False
    method = 'GET'
    status_code = 200
    verb = 'get'


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic

from .apigroupversionlist import APIGroupVersionList


class APIDescriptor(pydantic.BaseModel):
    host: str = pydantic.Field(
        default=...
    )

    kind: Literal['APIDescriptor'] = pydantic.Field(
        default='APIDescriptor'
    )

    groups: list[APIGroupVersionList] = pydantic.Field(
        default_factory=list
    )

    index: dict[str, APIGroupVersionList] = pydantic.Field(
        default_factory=dict,
        exclude=True
    )

    def model_post_init(self, _: Any) -> None:
        self.index = {g.group: g for g in self.groups}

    def add(self, group: APIGroupVersionList) -> None:
        if not self.index.get(group.group):
            self.index[group.group] = group
            self.groups.append(group)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from ..objectmeta import ObjectMeta
from .base import BaseNamespace
from .namespacespec import HierarchicalNamespaceSpec


class Project(BaseNamespace[ObjectMeta[str]], kind='Namespace', type='webiam.io/project'):
    model_config = {'populate_by_name': True}

    spec: HierarchicalNamespaceSpec[Literal['organization', 'folder'], int] = pydantic.Field(
        default_factory=HierarchicalNamespaceSpec,
        description="Defines the behavior of the `Namespace`."
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import logging
from typing import Any

import httpx


__all__: list[str] = [
    'AsyncClient',
]


class AsyncClient(httpx.AsyncClient):
    logger: logging.Logger = logging.getLogger('headless')

    @property
    def domain(self) -> str:
        return self.base_url.netloc.decode()

    def can_retry_timeout(self, request: httpx.Request):
        return request.method in {'GET', 'HEAD', 'OPTIONS'}

    async def request(self, *args: Any, **kwargs: Any) -> httpx.Response:
        attempts: int = 0
        delay: int = 5
        while True:
            attempts += 1
            must_log = attempts % 3 == 0
            try:
                response = await super().request(*args, **kwargs)
                response.raise_for_status()
                break
            except httpx.ConnectError as e:
                # Assumption is that there is no internet connection.
                if must_log:
                    self.logger.warning("Network issues prevent request (url: %s)", e.request.url)
                await asyncio.sleep(delay)
                continue
            except httpx.ConnectTimeout as e:
                # A connection timeout may occur due to intermediary issues, or
                # a booting application that has not bound its port yet.
                if must_log:
                    self.logger.warning("Connection timeout (url: %s)", e.request.url)
                await asyncio.sleep(delay)
                continue
            except httpx.ReadTimeout as e:
                # A read timeout means that we've succesfully connected, but there
                # was a timeout reading the response. This should only be retried on
                # safe HTTP methods, because we do not know what actually caused the
                # timeout, thus any destructive operations may actually have been
                # completed succesfully.
                if not self.can_retry_timeout(e.request):
                    raise
                await asyncio.sleep(delay)
                continue
            except httpx.HTTPStatusError as e:
                if e.response.status_code != 429:
                    response = e.response
                    break
                if must_log:
                    self.logger.warning("Request was rate-limited (url: %s)", e.response.url)
                await asyncio.sleep(delay)
        return response


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from canonical.ext.iam.protocols import IResourceAuthorizationContext


async def f(request: fastapi.Request):
    return await getattr(request.state, 'resource_context')


RequestAuthorizationContext: TypeAlias = Annotated[
    IResourceAuthorizationContext,
    fastapi.Depends(f)
]


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic

from .objectreference import ObjectReference


__all__: list[str] = [
    'NamespacedObjectReference'
]

G = TypeVar('G')
N = TypeVar('N')
K = TypeVar('K')
T = TypeVar('T')


class NamespacedObjectReference(ObjectReference[G, K, T], Generic[G, K, T, N]):
    model_config = {'populate_by_name': True}

    namespace: N = pydantic.Field(
        default=...,
        description=(
            "The `.metadata.namespace` of of the referent. Cannot be "
            "updated."
        ),
        frozen=True
    )


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .apidescriptor import APIDescriptor
from .apigroupversionlist import APIGroupVersionList
from .apiresourcelist import APIResource
from .apiresourcelist import APIResourceList
from .error import Error
from .iresourcequery import IResourceQuery
from .iresourcerepository import IResourceRepository
from .listbase import ListBase
from .namespace import Namespace
from .namespacedobjectmeta import NamespacedObjectMeta
from .namespacedobjectreference import NamespacedObjectReference
from .objectmeta import ObjectMeta
from .objectreference import ObjectReference
from .primarykey import PrimaryKey
from .resource import M as ObjectMetaType
from .resource import Resource
from .resourceinspector import ResourceInspector
from .resourcemetadata import ResourceMetadata as ResourceMeta
from .resourceserverlist import ResourceServerList
from .rootresource import ResourceType
from .rootresource import ResourceTypeVar
from .rootresource import RootResource
from .transientmeta import TransientMeta


__all__: list[str] = [
    'APIDescriptor',
    'APIGroupVersionList',
    'APIResource',
    'APIResourceList',
    'Error',
    'IResourceQuery',
    'IResourceRepository',
    'ListBase',
    'Namespace',
    'NamespacedObjectMeta',
    'NamespacedObjectReference',
    'ObjectMeta',
    'ObjectMetaType',
    'ObjectReference',
    'PrimaryKey',
    'Resource',
    'ResourceInspector',
    'ResourceMeta',
    'ResourceServerList',
    'ResourceType',
    'ResourceTypeVar',
    'RootResource',
    'TransientMeta',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import overload
from typing import Any
from typing import NoReturn
from typing import TypeVar

import httpx
import pydantic
from fastapi.exceptions import HTTPException

from canonical.ext.iam import PermissionQuery
from canonical.ext.iam import PermissionSet
from canonical.ext.iam.types import Permission
from canonical.ext.resource import APIDescriptor
from canonical.ext.resource import APIResource
from canonical.ext.resource import APIResourceList
from canonical.ext.resource import ResourceType
from canonical.ext.httpx import AsyncClient
from canonical.ext.resource import Error


R = TypeVar('R', bound=ResourceType)
T = TypeVar('T', bound=pydantic.BaseModel)


class ResourceClient(AsyncClient):
    _supported_resources: dict[str, APIResourceList] = {}
    logger: logging.Logger = logging.getLogger('uvicorn')

    def can_retry_timeout(self, request: httpx.Request):
        return super().can_retry_timeout(request) or any([
            str.endswith(request.url.path, ':permissions')
        ])

    def construct_uri(
        self,
        group: str,
        kind: str,
        name: str | None = None,
        namespace: str | None = None,
        version: str | None = None
    ):
        version = version or 'v1'
        url = f'{version}/{kind}'
        if group:
            url = f'{group}/{url}'
        if namespace is not None:
            url = f'/{group}/{version}/namespaces/{namespace}/{kind}'
        if name is not None:
            url = f'{url}/{name}'
        return url

    def is_discovered(self, api_version: str) -> bool:
        return api_version in self._supported_resources

    def raise_for_status(self, response: httpx.Response) -> NoReturn | None:
        if response.status_code >= 400:
            error = Error.model_validate_json(response.content)
            raise HTTPException(
                status_code=error.data.status_code,
                detail=error.data.detail
            )

    @overload
    def resource_factory(self, model: None, response: httpx.Response) -> dict[str, Any]:
        ...

    @overload
    def resource_factory(self, model: type[T], response: httpx.Response) -> T:
        ...

    def resource_factory(self, model: type[T] | None, response: httpx.Response) -> T | dict[str, Any]:
        return (
            response.json()
            if not model else
            model.model_validate_json(response.content)
        )

    async def apply(self, obj: dict[str, Any]) -> None:
        """Applies the given specification to the system."""
        metadata: dict[str, Any] | Any = obj.get('metadata')
        if metadata is None or not isinstance(metadata, dict):
            raise TypeError(
                "The metadata field is required and must be a "
                "dictionary."
            )
        name = metadata.get('name')
        generate = bool(metadata.get('generateName'))
        ns = metadata.get('namespace')
        if (name is None or not isinstance(name, (str, int))) and not generate:
            raise TypeError(
                "The .metadata.name field is required and must be a "
                "string."
            )
        resource, api_version = await self.discover_object(obj)
        if (ns is None or not isinstance(ns, str)) and resource.namespaced:
            raise TypeError(
                "The metadata.namespace field is required and must be a "
                "string."
            )
        method = self.create
        if name:
            method = self.replace
            exists = await self.exists(
                resource.path(
                    api_version=api_version,
                    name=name,
                    namespace=ns
                )
            )
            if not exists:
                method = self.create
            result = await method(resource, api_version, ns, name, obj)
        else:
            result = await method(resource, api_version, ns, name, obj)
        return result

    async def create(
        self,
        resource: APIResource,
        api_version: str,
        namespace: str | None,
        name: str | int | None,
        obj: dict[str, Any]
    ) -> None:
        response = await self.post(
            url=resource.path(api_version, namespace=namespace),
            json=obj
        )
        response.raise_for_status()

    async def discover_server(self, server: str) -> APIDescriptor:
        response = await self.get(url=server)
        self.raise_for_status(response)
        return APIDescriptor.model_validate_json(response.content)

    async def discover_object(self, obj: dict[str, Any]):
        api_version: str | None = obj.get('apiVersion')
        if api_version is None:
            raise TypeError("The apiVersion field is required.")
        kind = obj.get('kind')
        if kind is None:
            raise TypeError("The kind field is required.")
        if not self.is_discovered(api_version):
            response = await self.get(
                url=api_version,
                headers={'Accept': 'application/json'}
            )
            response.raise_for_status()
            self._supported_resources[api_version] = APIResourceList\
                .model_validate_json(response.content)

        return self._supported_resources[api_version].get(kind), api_version

    async def exists(self, url: str) -> bool:
        response = await self.head(
            url=url,
            follow_redirects=True
        )
        if response.status_code not in {200, 404}:
            response.raise_for_status()
        return response.status_code == 200

    async def permissions(
        self,
        namespace: str | None,
        permission: str,
        *permissions: str,
        **kwargs: Any
    ) -> set[Permission]:
        """Retrieve the permissions that the client has on the given
        resource.
        """
        query = PermissionQuery.factory({
            'permissions': {permission, *permissions}
        })
        headers: dict[str, str] = kwargs.setdefault('headers', {})
        headers.setdefault('Accept', 'application/json')
        response = await self.post(
            headers=headers,
            url=f"{self.construct_uri('', 'namespaces', namespace)}:permissions",
            json=query.model_dump(by_alias=True, mode='json')
        )
        if response.status_code >= 400:
            self.logger.warning(
                "Permissions endpoint responded with %s (endpoint: %s)",
                response.status_code, response.url
            )
            result = PermissionSet(granted=set())
        else:
            result = PermissionSet.model_validate_json(response.content)
        return {Permission(x) for x in result.granted}

    @overload
    async def request_resource(self, model: None, **kwargs: Any) -> dict[str, Any]:
        ...

    @overload
    async def request_resource(self, model: type[T], **kwargs: Any) -> T:
        ...

    async def request_resource(self, model: type[T] | None, method: str, **kwargs: Any) -> T | dict[str, Any]:
        headers = kwargs.setdefault('headers', {})
        headers['Accept'] = 'application/json'
        response = await self.request(method=method, **kwargs)
        self.raise_for_status(response)
        resource = response.json()
        if model is not None:
            resource = self.resource_factory(model, response)
        return resource

    async def retrieve(
        self,
        group: str,
        kind: str,
        name: str,
        model: type[T] | None = None,
        namespace: str | None = None,
        version: str | None = None,
    ) -> dict[str, Any] | T:
        return await self.request_resource(
            model=model,
            method='GET',
            url=self.construct_uri(group, kind, name, namespace, version)
        )

    async def replace(
        self,
        resource: APIResource,
        api_version: str,
        namespace: str | None,
        name: str | int,
        obj: dict[str, Any]
    ) -> None:
        response = await self.put(
            url=resource.path(api_version, name=name, namespace=namespace),
            json=obj
        )
        response.raise_for_status()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Annotated
from typing import TypeVar

import fastapi
import fastapi.params

from canonical.exceptions import ProgrammingError


__all__: list[str] = [
    'request_state'
]


UNSET = object()
T = TypeVar('T')


def request_state(
    name: str,
    annotation: type[T],
    required: bool = True,
    default: object = UNSET
) -> T:
    if not bool(default) ^ bool(required):
        raise TypeError(
            "The `required` and `default` parameters are mutually "
            "exclusive."
        )

    def f(request: fastapi.Request):
        v = getattr(request, name, default)
        if v == UNSET and required:
            raise ProgrammingError(
                f"Dependency requires request.state.{name} "
                "to be present."
            )
        setattr(request.state, name, v)
        return v

    return cast(T, Annotated[annotation, fastapi.Depends(f)])




# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal
from typing import Union

import pydantic


JSONPatchOperationType = Literal[
    'add',
    'replace',
    'remove',
    'move',
    'copy',
    'test'
]


class JSONPatchValueOperation(pydantic.BaseModel):
    op: Literal['add', 'replace', 'test'] = pydantic.Field(
        default=...,
        description=(
            "The operation to be applied to the target JSON document. "
            "Must be `add`, `replace` or `test`."
        )
    )

    path: str = pydantic.Field(
        default=...,
        description=(
            "Path to apply the patch to in the target JSON document. "
            "Must start with a slash."
        )
    )


class JSONPatchMoveOperation(pydantic.BaseModel):
    op: Literal['move'] = pydantic.Field(
        default=...,
        description=(
            "The operation to be applied to the target JSON document. "
            "Must be `move`."
        )
    )

    path: str = pydantic.Field(
        default=...,
        description=(
            "Path in the target document to move `from`."
        )
    )

    source: str = pydantic.Field(
        default=...,
        alias='from',
        description=(
            "Source path in the target JSON document."
        )
    )



class JSONPatchCopyOperation(pydantic.BaseModel):
    op: Literal['copy'] = pydantic.Field(
        default=...,
        description=(
            "The operation to be applied to the target JSON document. "
            "Must be `copy`."
        )
    )

    path: str = pydantic.Field(
        default=...,
        description=(
            "Path in the target document to copy `from`."
        )
    )

    source: str = pydantic.Field(
        default=...,
        alias='from',
        description=(
            "Source path in the target JSON document."
        )
    )


class JSONPatchRemoveOperation(pydantic.BaseModel):
    op: Literal['remove'] = pydantic.Field(
        default=...,
        description=(
            "The operation to be applied to the target JSON document. "
            "Must be `remove`."
        )
    )

    path: str = pydantic.Field(
        default=...,
        description=(
            "Path to remove from the target JSON document."
        )
    )


JSONPatchType = Union[
    JSONPatchValueOperation,
    JSONPatchCopyOperation,
    JSONPatchMoveOperation,
    JSONPatchRemoveOperation
]